const fs = require('fs'),
      sys = require('sys'),
      events = require('events');

/**
 * Constructor
 */
function FakeDB(path) {

  // call the parent
  events.EventEmitter.call(this);

  var self = this;

  this._hash = {};
  this._list = [];
  this._operations = [];
  this._logging = false;
  this._flushing = false;

  // data index/map
  this._index = {};

  this.path = path;
  
  this._rstr = fs.createReadStream(this.path, {
    flags: 'r',
    encoding: 'utf-8',
    mode: 0666
  });
  
  this._wstr = fs.createWriteStream(this.path, {
    flags: 'a+',
    encoding: 'utf-8',
    mode: 0666
  });

  this._wstr.on('drain', function () {
    self._flushing = false;
    if (self._operations.length) {
      // if we have more stuff to write
      self._flush();
    } else {
      self.emit('drain');
    }
  });

  this._load();
}

/**
 * We'll be emitting 'load', 'drain' and 'error'
 */
sys.inherits(FakeDB, events.EventEmitter);

/**
 * Adds operation to log
 */
FakeDB.prototype._log = function (entry) {
  if (this._logging) {
    this._operations.push(entry);
  }
};

/**
 * Load data from a db file
 */
FakeDB.prototype._load = function () {

  var buffer = '',
      stream = null,
      self = this;
      
  this._rstr.on('error', function (err) {
    if (err.code == 'ENOENT') {
      self._logging = true;
      self.emit('load');
    }
  });

  this._rstr.on('data', function (chunk) {
    buffer += chunk;
  });

  this._rstr.on('end', function () {

    var lines = buffer.split('\n');
    lines.forEach(function (row) {
      if (row.length) {
        try {
          row = JSON.parse(row);
          if ('key' in row && 'value' in row) {
            self.set(row.key, row.value);
          } else {
            throw new Error('Invalid record encountered');
          }
        } catch (e) {
          self.emit('error', e);
        }
      }
    });

    // log all operations after loading data
    self._logging = true;

    self.emit('load');
  });

};

/**
 * Flush the log queue into the db file
 */
FakeDB.prototype._flush = function () {

  if (this._flushing || !this._operations.length) {
    return;
  }

  this._flushing = true;

  var data = null,
      self = this;

  while (data = this._operations.shift()) {
    this._wstr.write(JSON.stringify(data) + '\n', function (err) {
      if (err) {
        self.emit('error', new Error('Failed to append to log file.'));
      }
    });
  }
};

/**
 * Set a key in the database
 */
FakeDB.prototype.set = function (key, value) {

  if (value !== null) {
    if (!(key in this._hash)) {
      this._list.push(key);
      this._updateIndex(key, value);
    }
    this._hash[key] = value;
    if (this._logging) {
      this._log({ key: key, value: value });
      this._flush();
    }
  } else {
    this.remove(key);
  }
};

/**
 * Remove an object
 */
FakeDB.prototype.remove = function (key) {
  if (key in this._hash) {
    this._removeFromIndex(key);
    delete this._list.splice(this._list.indexOf(key), 1);
    delete this._hash[key];
    if (this._logging) {
      this._log({ key: key, value: null });
      this._flush();
    }
  }
};

/**
 * Get a stored key
 */
FakeDB.prototype.get = function (key, cb) {
  if (key in this._hash) {
    cb.call(undefined, key, this._hash[key]);
  } else {
    cb.call(undefined, null, null);
  }
};

/**
 * Get all records
 */
FakeDB.prototype.all = function (opt, cb) {

  var self = this,
      result = [];

  if (typeof opt == 'function') {
    cb = opt;
    opt = { offset: 0, limit: 0 };
  }
  
  var i = (opt.offset > this._list.length ? this._list.length : opt.offset),
      n = (opt.limit > 0 ? opt.limit : this._list.length);

  this._list.slice(i, n).forEach(function (key) {
    result.push(self._hash[key]);
  });
  
  cb.call(undefined, result);

};

/**
 * First record
 */
FakeDB.prototype.first = function (cb) {
  if (this._list.length) {
    cb.call(undefined, this._list[0], this._hash[this._list[0]]);
  }
};

/**
 * Last record
 */
FakeDB.prototype.last = function (cb) {
  if (this._list.length) {
    var k = this._list[this._list.length - 1];
    var v = this._hash[k];
    cb.call(undefined, k, v);
  }
};

/**
 * Invoke cb on all records that cause checker() to return true
 */
FakeDB.prototype.filter = function (checker, cb) {
  for (var i = 0, n = this._list.length, o = null; i < n; ++i) {
    o = this._hash[this._list[i]];
    if (checker.call(undefined, o)) {
      cb.call(undefined, o);
    }
  }
};

/**
 * Search and return the first one
 */
FakeDB.prototype.findOne = function (qcb, cb) {
  var result = null;
  for (var i = 0, n = this._list.length, o = null; i < n; ++i) {
    o = this._hash[this._list[i]];
    if (qcb.call(undefined, o)) {
      cb.call(undefined, o);
      return;
    }
  }
};

/**
 * Index creation
 */
FakeDB.prototype.index = function (field) {

  if (!(field in this._index)) {
    this._index[field] = { vals: [], keys: [] };
  }

  for (var key in this._hash) {
    this._addToIndex(field, key, this._hash[key]);
  }
};

/**
 * Get position of a record from the index
 */
FakeDB.prototype._getIndexPosition = function (field, value) {
  
  if (!(field in this._index)) {
    return { position: -1, found: false, direction: 0 };
  }

  var keys = this._index[field].keys,
      vals = this._index[field].vals,
      l = 0,
      h = vals.length - 1,
      i = 0,
      c = 0,
      d = 0,
      s = [],
      f = false;

  value = value.toLowerCase();
  while (l <= h) {

    i = Math.floor((l + h) / 2);
    c = vals[i].toLowerCase().localeCompare(value);

    if (c != 0) {
      if (c < 0) {
        l = i + 1;
        d = 1;
      } else {
        h = i - 1;
        d = 0;
      }
      continue;
    }

    f = true;
    break;
  }

  return { position: i, found: f, direction: d };
};

/**
 * Update indices with record
 */
FakeDB.prototype._updateIndex = function (key, record) {
  for (var field in this._index) {
    this._addToIndex(field, key, record);
  }
};

/**
 * Add index entry
 */
FakeDB.prototype._addToIndex = function (field, key, record) {
  var p = null;
  if (field in record) {
    p = this._getIndexPosition(field, record[field]);
    if (p.position >= 0) {
      if (p.found) {
        this._index[field].keys[p.position].push(key);
      } else {
        this._index[field].vals.splice(p.position + p.direction, 0, record[field]);
        this._index[field].keys.splice(p.position + p.direction, 0, [key]);
      }
    }
  }
};

/**
 * Remove index entry
 */
FakeDB.prototype._removeFromIndex = function (key) {
  
  var p = null,
      o = this._hash[key],
      v = null;
  
  for (var field in this._index) {
    if (field in o) {
      p = this._getIndexPosition(field, o[field]);
      if (p.found) {
        v = this._index[field].keys[p.position];
        v.splice(v.indexOf(key), 1);
      }
    }
  }
};

/**
 * Indexed search
 */
FakeDB.prototype.search = function (query, opt, cb) {

  var result = [],
      cursor = null,
      keys = null,
      queue = [];

  if (typeof opt == 'function') {
    cb = opt;
    opt = { offset: 0, limit: 0 };
  }

  for (var field in query) {
    cursor = this._getIndexPosition(field, query[field]);
    if (cursor.found) {
      keys = this._index[field].keys[cursor.position];
      for (var i = 0, n = keys.length; i < n; ++i) {
        if (queue.indexOf(keys[i]) < 0) {
          queue.push(keys[i]);
          result.push(this._hash[keys[i]]);
        }
      }
    }
  }

  var i = (opt.offset > this._list.length ? this._list.length : opt.offset),
      n = (opt.limit > 0 ? opt.limit : this._list.length);

  cb.call(undefined, result.slice(i, i + n));
};

/**
 * Returns number of records in DB
 */
FakeDB.prototype.count = function () {
  return this._list.length;
};

exports.init = function (path) {
  return new FakeDB(path);
};
